"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.stringify = void 0;
const types_1 = require("./types");
const isReactElement_1 = require("./utils/isReactElement");
const isCallable_1 = require("./utils/isCallable");
const isObject_1 = require("./utils/isObject");
function stringify(value, { forbidReactElements, space, valueName = 'value', sortObjectKeys } = {}) {
    const path = [];
    // The only error `JSON.stringify()` can throw is `TypeError "cyclic object value"`.
    // - https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/JSON/stringify#exceptions
    // - This means we have total of 3 possible errors while serializing:
    //    - Cyclic references
    //    - Functions
    //    - React elements
    const serializer = (val) => JSON.stringify(val, replacer, space);
    return serializer(value);
    function replacer(key, value) {
        if (key !== '') {
            path.push(key);
        }
        if (forbidReactElements && (0, isReactElement_1.isReactElement)(value)) {
            throw new Error(genErrMsg('React element'));
        }
        if ((0, isCallable_1.isCallable)(value)) {
            const functionName = value.name;
            throw new Error(genErrMsg('function', path.length === 0 ? functionName : undefined));
        }
        const valueOriginal = this[key];
        for (const { is, serialize } of types_1.types.slice().reverse()) {
            if (is(valueOriginal)) {
                //@ts-ignore
                return serialize(valueOriginal, serializer);
            }
        }
        if (sortObjectKeys && (0, isObject_1.isObject)(value)) {
            const copy = {};
            Object.keys(value)
                .sort()
                .forEach((key) => {
                copy[key] = value[key];
            });
            value = copy;
        }
        return value;
    }
    function genErrMsg(valueType, valName) {
        const name = valName ? ' `' + valName + '`' : '';
        const location = path.length === 0 ? '' : ` ${name ? 'at ' : ''}\`${valueName}[${path.map((p) => `'${p}'`).join('][')}]\``;
        const fallback = name === '' && location === '' ? ` ${valueName}` : '';
        return `Cannot serialize${name}${location}${fallback} because it is a ${valueType} (https://github.com/brillout/json-s)`;
    }
}
exports.stringify = stringify;
